
#include <xapian.h>
#include "fts-xapian-plugin.h"

#define XAPIAN_FILE_PREFIX "xapian-indexes"

struct xapian_fts_backend {
        struct fts_backend backend;
        char path[1000];

        struct mailbox *box;
        char guid[200];

        char db[1000];

        unsigned int partial,full;
} ;

struct xapian_fts_backend_update_context
{
        struct fts_backend_update_context ctx;
        char tbi_field[100];
        bool tbi_isfield;
        uint32_t tbi_uid;
        bool failed;
};

#include "fts-backend-xapian-functions.cpp"


static struct fts_backend *fts_backend_xapian_alloc(void)
{
        struct xapian_fts_backend *backend;

        backend = i_new(struct xapian_fts_backend, 1);
        backend->backend = fts_backend_xapian;
        return &backend->backend;
}

static int fts_backend_xapian_init(struct fts_backend *_backend, const char **error_r)
{
	struct xapian_fts_backend *backend =
		(struct xapian_fts_backend *)_backend;
	const char *const *tmp, *env;
	unsigned int len;

	env = mail_user_plugin_getenv(_backend->ns->user, "fts_xapian");
	if (env == NULL)
		return 0;

	for (tmp = t_strsplit_spaces(env, " "); *tmp != NULL; tmp++) 
	{
        	if (str_begins(*tmp, "partial=") && (str_to_uint(*tmp + 8, &len)>=0))
		{
			backend->partial=len;
		}	
        	else if (str_begins(*tmp, "full=") && (str_to_uint(*tmp + 5, &len)>=0))
		{
			backend->full=len;
		}
        	else 
		{
            		i_error("fts_xapian: Invalid setting: %s", *tmp);
            		return -1;
        	}
    	}
    	if(backend->partial<1) 
    	{
        	i_error("fts_xapian: 'partial' can not be null (try partial=3)");
        	return -1;
    	}
    	if(backend->full<1) 
    	{
        	i_error("fts_xapian: 'full' can not be null (try full=25)");
        	return -1;
    	}
    	if(backend->partial > backend->full) 
    	{
        	i_error("fts_xapian: 'full' must be equal or greater than 'partial'");
        	return -1;
    	}
    	if(backend->full > 50) 
    	{
        	i_error("fts_xapian: 'full' above 50 is not realistic");
        	return -1;
    	}

	return 0;
}

static void fts_backend_xapian_deinit(struct fts_backend *_backend)
{
	struct xapian_fts_backend *backend =
		(struct xapian_fts_backend *)_backend;
	// DEINIT CODE
	// 	
	//
	fts_backend_xapian_unset_box(backend);
	i_free(backend);
}


static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend,
			       struct mailbox *box, uint32_t *last_uid_r)
{
	struct xapian_fts_backend *backend =
		(struct xapian_fts_backend *)_backend;

    	if(strlen(backend->db)<1) return -1;
	
    	try
	{
		Xapian::Database db(backend->db);

    		Xapian::docid docid=db.get_lastdocid();
    		Xapian::Document doc = db.get_document(docid);
    		return atoi(doc.get_value(1).c_str());
	}
	catch(Xapian::Error e)
	{
		i_error(e.get_msg().c_str());
	}
	return -1;
}


static struct fts_backend_update_context * fts_backend_xapian_update_init(struct fts_backend *_backend)
{
	struct xapian_fts_backend_update_context *ctx;

	ctx = i_new(struct xapian_fts_backend_update_context, 1);
	ctx->ctx.backend = _backend;
    	ctx->failed=false;
	return &ctx->ctx;
}

static int fts_backend_xapian_update_deinit(struct fts_backend_update_context *_ctx)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;

	int ret = ctx->failed ? -1 : 0;

	i_free(ctx);
	return ret;
}

static void fts_backend_xapian_update_set_mailbox(struct fts_backend_update_context *_ctx, struct mailbox *box)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;
	struct xapian_fts_backend *backend =
		(struct xapian_fts_backend *)ctx->ctx.backend;

    	if(fts_backend_xapian_set_box(backend, box) < 0)
		ctx->failed = TRUE;
}

static void fts_backend_xapian_update_expunge(struct fts_backend_update_context *_ctx, uint32_t uid)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;
	struct xapian_fts_backend *backend =
		(struct xapian_fts_backend *)ctx->ctx.backend;

    	if(strlen(backend->db)<1) return;

    	try
	{
		XQuerySet xq;
		char u[1000]; sprintf(u,"%i",uid);
		xq.add("uid",u);
		XResultSet * result=fts_backend_xapian_query(backend->db,&xq);
    
    		if(result->size>0)
    		{
			Xapian::WritableDatabase db(backend->db,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK);
			Xapian::docid docid = result->data[0]; 
        		db.delete_document(docid);
    		}
		free(result);
	}
	catch(Xapian::Error e)
	{
		i_error(e.get_msg().c_str());
	}
}

static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_context *_ctx, const struct fts_backend_build_key *key)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;

	if (ctx->failed)
		return FALSE;

	switch (key->type) 
    	{
    		case FTS_BACKEND_BUILD_KEY_HDR:
	    	case FTS_BACKEND_BUILD_KEY_MIME_HDR:
            		ctx->tbi_isfield=true;
            		if(strcmp(key->hdr_name,"Subject")==0)
            		{
                		ctx->tbi_isfield=false;
            		}
            		sprintf(ctx->tbi_field,"%s",key->hdr_name);
            		ctx->tbi_uid=key->uid;
            		break;
        	case FTS_BACKEND_BUILD_KEY_BODY_PART:
            		sprintf(ctx->tbi_field,"%s","BODY");
            		ctx->tbi_isfield=false;
            		ctx->tbi_uid=key->uid;
            		break;
	    	case FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY:
		    	i_unreached();
	}
	return TRUE;
}

static void fts_backend_xapian_update_unset_build_key(struct fts_backend_update_context *_ctx)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;

    	ctx->tbi_uid=0;
}

static int fts_backend_xapian_update_build_more(struct fts_backend_update_context *_ctx, const char *data, size_t size)
{
	struct xapian_fts_backend_update_context *ctx =
		(struct xapian_fts_backend_update_context *)_ctx;
	struct xapian_fts_backend *backend =
                (struct xapian_fts_backend *) ctx->ctx.backend;

	if(strlen(backend->db)<1) return -1;

    	if(ctx->tbi_uid<1) return -1;

    	if(ctx->tbi_isfield)
    	{
        	if(!fts_backend_xapian_index_hdr(backend->db,ctx->tbi_uid,ctx->tbi_field,data,backend->partial,backend->full))
        	{
            		ctx->failed=true;
            		return -1;
        	}
    	}
    	else
    	{
        	if(!fts_backend_xapian_index_text(backend->db,ctx->tbi_uid,ctx->tbi_field,data))
        	{
            		ctx->failed=true;
            		return -1;
        	}
    	}
    	return 0;
}

static int fts_backend_xapian_rescan(struct fts_backend *_backend)
{
	struct xapian_fts_backend *backend =
                (struct xapian_fts_backend *) _backend;

	if(strlen(backend->db)<1) return -1;
  
	try
        {
		Xapian::WritableDatabase db(backend->db,Xapian::DB_CREATE_OR_OVERWRITE | Xapian::DB_RETRY_LOCK);
		return 0;
	}
	catch(Xapian::Error e)
	{
		i_error(e.get_msg().c_str());
	}
	return -1;
}


static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbox *box, struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_result *result)
{
	struct xapian_fts_backend *backend =
                (struct xapian_fts_backend *) _backend;

	if(fts_backend_xapian_set_box(backend, box)<0)
		return -1;

	XQuerySet * set = new XQuerySet();

	set->is_and = (flags & FTS_LOOKUP_FLAG_AND_ARGS) != 0;

	for (; args != NULL; args = args->next) 
	{
		set->add(args->hdr_field_name,args->value.str);
	}
	free(set);
	return 0;
}

 

struct fts_backend fts_backend_xapian = {
	.name = "xapian",
	.flags = FTS_BACKEND_FLAG_NORMALIZE_INPUT,
	{
		fts_backend_xapian_alloc,
		fts_backend_xapian_init,
		fts_backend_xapian_deinit,
		fts_backend_xapian_get_last_uid,
		fts_backend_xapian_update_init,
		fts_backend_xapian_update_deinit,
		fts_backend_xapian_update_set_mailbox,
		fts_backend_xapian_update_expunge,
		fts_backend_xapian_update_set_build_key,
		fts_backend_xapian_update_unset_build_key,
		fts_backend_xapian_update_build_more,
		NULL,
		fts_backend_xapian_rescan,
		NULL,
		fts_backend_default_can_lookup,
		fts_backend_xapian_lookup,
		NULL,
        	NULL
	}
};

